@@ -19,7 +19,7 @@
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 
-# This generates libvirt_c.c (the core of the bindings).  You don't
+# This generates libvirt_generated.c (the core of the bindings).  You don't
 # need to run this program unless you are extending the bindings
 # themselves (eg. because libvirt has been extended).
 #
@@ -91,6 +91,7 @@ my @functions = (
     { name => "virDomainCreate", sig => "dom : unit" },
     { name => "virDomainAttachDevice", sig => "dom, string : unit" },
     { name => "virDomainDetachDevice", sig => "dom, string : unit" },
+    { name => "virDomainDetachDeviceFlags", sig => "dom, string, unsigned : unit" },
     { name => "virDomainGetAutostart", sig => "dom : bool" },
     { name => "virDomainSetAutostart", sig => "dom, bool : unit" },
 
@@ -199,7 +200,7 @@ my @unimplemented = (
 
 # Open the output file.
 
-my $filename = "libvirt_c.c";
+my $filename = "libvirt_generated.c";
 open F, ">$filename" or die "$filename: $!";
 
 # Write the prologue.
@@ -232,28 +233,11 @@ print F <<'END';
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
  */
 
-#include "config.h"
+#include "libvirt_c.h"
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-
-#include <libvirt/libvirt.h>
-#include <libvirt/virterror.h>
-
-#include <caml/config.h>
-#include <caml/alloc.h>
-#include <caml/callback.h>
-#include <caml/custom.h>
-#include <caml/fail.h>
-#include <caml/memory.h>
-#include <caml/misc.h>
-#include <caml/mlvalues.h>
-#include <caml/signals.h>
-
-#include "libvirt_c_prologue.c"
-
-#include "libvirt_c_oneoffs.c"
+#ifdef __GNUC__
+#pragma GCC diagnostic ignored "-Wmissing-prototypes"
+#endif
 
 END
 
@@ -457,7 +441,7 @@ sub gen_c_code
 
   /* UUIDs are byte arrays with a fixed length. */
   rv = caml_alloc_string (VIR_UUID_BUFLEN);
-  memcpy (String_val (rv), uuid, VIR_UUID_BUFLEN);
+  memcpy (Bytes_val (rv), uuid, VIR_UUID_BUFLEN);
   CAMLreturn (rv);
 "
     } elsif ($sig =~ /^(\w+) : uuid string$/) {
@@ -610,7 +594,7 @@ sub gen_c_code
     } elsif ($sig =~ /^(\w+), string : unit$/) {
 	"\
   " . gen_unpack_args ($1) . "
-  char *str = String_val (strv);
+  const char *str = String_val (strv);
   int r;
 
   NONBLOCKING (r = $c_name ($1, str));
@@ -622,7 +606,7 @@ sub gen_c_code
 	"\
   CAMLlocal1 (rv);
   " . gen_unpack_args ($1) . "
-  char *str = String_val (strv);
+  const char *str = String_val (strv);
   int r;
 
   NONBLOCKING (r = $c_name ($1, str, 0));
@@ -635,7 +619,7 @@ sub gen_c_code
 	"\
   CAMLlocal1 (rv);
   " . gen_unpack_args ($1) . "
-  char *str = String_val (strv);
+  const char *str = String_val (strv);
   $c_ret_type r;
 
   NONBLOCKING (r = $c_name ($1, str));
@@ -650,7 +634,7 @@ sub gen_c_code
 	"\
   CAMLlocal1 (rv);
   " . gen_unpack_args ($1) . "
-  char *str = String_val (strv);
+  const char *str = String_val (strv);
   $c_ret_type r;
 
   NONBLOCKING (r = $c_name ($1, str, 0));
@@ -659,13 +643,25 @@ sub gen_c_code
   " . gen_pack_result ($2) . "
 
   CAMLreturn (rv);
+"
+    } elsif ($sig =~ /^(\w+), string, unsigned : unit$/) {
+	"\
+  " . gen_unpack_args ($1) . "
+  const char *str = String_val (strv);
+  unsigned int u = Int_val (uv);
+  int r;
+
+  NONBLOCKING (r = $c_name ($1, str, u));
+  CHECK_ERROR (r == -1, \"$c_name\");
+
+  CAMLreturn (Val_unit);
 "
     } elsif ($sig =~ /^(\w+), string, unsigned : (\w+)$/) {
 	my $c_ret_type = short_name_to_c_type ($2);
 	"\
   CAMLlocal1 (rv);
   " . gen_unpack_args ($1) . "
-  char *str = String_val (strv);
+  const char *str = String_val (strv);
   unsigned int u = Int_val (uv);
   $c_ret_type r;
 
@@ -752,7 +748,7 @@ sub gen_c_code
 	"\
   CAMLlocal2 (rv, connv);
   " . gen_unpack_args ($1) . "
-  char *str = String_val (strv);
+  const char *str = String_val (strv);
   $c_ret_type r;
 
   NONBLOCKING (r = $c_name ($1, str));
@@ -768,7 +764,7 @@ sub gen_c_code
 	"\
   CAMLlocal2 (rv, connv);
   " . gen_unpack_args ($1) . "
-  char *str = String_val (strv);
+  const char *str = String_val (strv);
   $c_ret_type r;
 
   NONBLOCKING (r = $c_name ($1, str, 0));
@@ -899,8 +895,6 @@ END
 # Write the epilogue.
 
 print F <<'END';
-#include "libvirt_c_epilogue.c"
-
 /* EOF */
 END
 
