@@ -517,6 +517,11 @@ sig
   | START_FORCE_BOOT                    (** Discard any managed save *)
   | START_VALIDATE                      (** Validate XML against schema *)
 
+  type domain_device_modify_flag =
+  | DEVICE_MODIFY_LIVE                  (** Affect running domain state *)
+  | DEVICE_MODIFY_CONFIG                (** Affect persistent domain state *)
+  | DEVICE_MODIFY_FORCE                 (** Affect running domain state forcedly *)
+
   type sched_param = string * sched_param_value
   and sched_param_value =
     | SchedFieldInt32 of int32 | SchedFieldUInt32 of int32
@@ -690,6 +695,9 @@ sig
     (** Attach a device (described by the device XML) to a domain. *)
   val detach_device : [>`W] t -> xml -> unit
     (** Detach a device (described by the device XML) from a domain. *)
+  val detach_device_flags : [>`W] t -> xml -> domain_device_modify_flag list -> unit
+    (** Detach a device (described by the device XML) from a domain, with the
+  possibility	to specify flags. *)
 
   val migrate : [>`W] t -> [>`W] Connect.t -> migrate_flag list ->
     ?dname:string -> ?uri:string -> ?bandwidth:int -> unit -> rw t
@@ -708,7 +716,7 @@ sig
   val interface_stats : [>`R] t -> string -> interface_stats
     (** Returns network interface stats. *)
 
-  val block_peek : [>`W] t -> string -> int64 -> int -> string -> int -> unit
+  val block_peek : [>`W] t -> string -> int64 -> int -> bytes -> int -> unit
     (** [block_peek dom path offset size buf boff] reads [size] bytes at
 	[offset] in the domain's [path] block device.
 
@@ -717,7 +725,7 @@ sig
 
 	See also {!max_peek}. *)
   val memory_peek : [>`W] t -> memory_flag list -> int64 -> int ->
-    string -> int -> unit
+    bytes -> int -> unit
     (** [memory_peek dom Virtual offset size] reads [size] bytes
 	at [offset] in the domain's virtual memory.
 
@@ -1058,6 +1066,11 @@ sig
     val to_string: t -> string
   end
 
+  module Device_removed : sig
+    type t = string
+
+    val to_string: t -> string
+  end
 
   type callback =
     | Lifecycle     of ([`R] Domain.t -> Lifecycle.t -> unit)
@@ -1075,6 +1088,7 @@ sig
     | PMSuspend     of ([`R] Domain.t -> PM_suspend.t -> unit)
     | BalloonChange of ([`R] Domain.t -> Balloon_change.t -> unit)
     | PMSuspendDisk of ([`R] Domain.t -> PM_suspend_disk.t -> unit)
+    | DeviceRemoved of ([`R] Domain.t -> Device_removed.t -> unit)
 
     (** type of a registered call back function *)
 
@@ -1484,8 +1498,9 @@ sig
     | VIR_ERR_INVALID_NETWORK_PORT
     | VIR_ERR_NETWORK_PORT_EXIST
     | VIR_ERR_NO_NETWORK_PORT
+    | VIR_ERR_NO_HOSTNAME
 	(* ^^ NB: If you add a variant you MUST edit
-	   libvirt_c_epilogue.c:MAX_VIR_* *)
+	   libvirt_c_common.c:MAX_VIR_* *)
     | VIR_ERR_UNKNOWN of int (** Other error, not handled with existing values. *)
 	(** See [<libvirt/virterror.h>] for meaning of these codes. *)
 
@@ -1563,8 +1578,9 @@ sig
     | VIR_FROM_FIREWALLD
     | VIR_FROM_DOMAIN_CHECKPOINT
     | VIR_FROM_TPM
+    | VIR_FROM_BPF
 	(* ^^ NB: If you add a variant you MUST edit
-	   libvirt_c_epilogue.c: MAX_VIR_* *)
+	   libvirt_c_common.c: MAX_VIR_* *)
     | VIR_FROM_UNKNOWN of int (** Other domain, not handled with existing values. *)
 	(** Subsystem / driver which produced the error. *)
 
@@ -1574,7 +1590,7 @@ sig
     | VIR_ERR_NONE
     | VIR_ERR_WARNING
     | VIR_ERR_ERROR
-	(* ^^ NB: If you add a variant you MUST edit libvirt_c.c: MAX_VIR_* *)
+	(* ^^ NB: If you add a variant you MUST edit libvirt_generated.c: MAX_VIR_* *)
     | VIR_ERR_UNKNOWN_LEVEL of int (** Other level, not handled with existing values. *)
 	(** No error, a warning or an error. *)
 
